# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import time
from machine import enable_irq, disable_irq
from micropython import const

from terkin import logging
from terkin.util import get_platform_info

log = logging.getLogger(__name__)


# Todo: Make this configurable.
#log.setLevel(logging.DEBUG)


class AbstractSystemSensor:

    def __init__(self, settings):
        self.settings = settings
        self.platform_info = get_platform_info()


class SystemMemoryFree(AbstractSystemSensor):
    """Read free memory in bytes."""

    def read(self):
        """ """
        import gc
        value = gc.mem_free()
        return {'system.memfree': value}


class SystemTemperature(AbstractSystemSensor):
    """
    Read the internal temperature of the MCU.
    
    - https://docs.micropython.org/en/latest/esp32/quickref.html#general-board-control
    - https://github.com/micropython/micropython-esp32/issues/33
    - https://github.com/micropython/micropython/pull/3933
    - https://github.com/micropython/micropython-esp32/pull/192
    - https://github.com/espressif/esp-idf/issues/146
    - https://forum.pycom.io/topic/2208/new-firmware-release-1-10-2-b1/4

    """

    def read(self):
        """ """
        import machine

        if not hasattr(machine, 'temperature'):
            raise NotImplementedError(
                'Reading the MCU core temperature is not implemented on this platform')

        rawvalue = machine.temperature()

        # Fahrenheit
        # 'system.temperature': 57.77778
        #temperature = (rawvalue - 32) / 1.8

        # Magic
        # 'system.temperature': 41.30435
        value = rawvalue * (44/23) + (-5034/23)

        reading = {'system.temperature': value}
        return reading


class SystemVoltage(AbstractSystemSensor):
    """
    Read the a voltage level by sampling the ADC on a pin connected
    to a voltage divider. As the Pycom expansion board is using
    Pin 16 for battery voltage, this is also used on other boards
    as kind of a convention.
    
    Implementation
    ==============
    Written by Dominik Kapusta <https://github.com/ayoy>. Thanks!
    Improved by Andreas Motl <https://github.com/amotl>.
    
    License
    =======
    The MIT License (MIT)
    
    Copyright (c) 2018 Dominik Kapusta
    
    - https://kapusta.cc/2018/02/02/air-quality-monitor-revisited/
    - https://github.com/ayoy/upython-aq-monitor/blob/lora/lib/adc.py
    
    Documentation
    =============
    - https://docs.pycom.io/firmwareapi/pycom/machine/adc
    - https://docs.pycom.io/tutorials/all/adc
    
    More resources
    ==============
    - https://forum.pycom.io/topic/3776/adc-use-to-measure-battery-level-vin-level
    - https://github.com/hiveeyes/terkin-datalogger/issues/5
    - https://community.hiveeyes.org/t/batterieuberwachung-voltage-divider-und-attenuation-fur-microypthon-firmware/2128


    """

    # How many times to sample the ADC for making a reading.
    adc_sample_count = const(1000)

    def __init__(self, settings):
        """
        Initialized ADC unit.
        """

        super().__init__(settings)

        # ADC Pin to sample from.
        self.pin = None

        # Main resistor value (R1).
        self.resistor_r1 = None

        # Resistor between input pin and ground (R2).
        self.resistor_r2 = None

        # Reference to platform ADC object.
        self.adc = None

        self.setup()

    def setup(self):
        """
        - Configure the appropriate resistor values for computing the voltage.
        - Setup ADC for sampling.
        """

        self.pin = self.settings.get('pin')
        self.resistor_r1 = self.settings.get('resistor_r1')
        self.resistor_r2 = self.settings.get('resistor_r2')
        self.adc_attenuation_db = self.settings.get('adc_attenuation_db', 6.0)
        self.reading_key = self.settings.get('type')

        assert type(self.pin) is str, 'VCC Error: Voltage divider ADC pin invalid'
        assert type(self.resistor_r1) is int, 'VCC Error: Voltage divider resistor value "resistor_r1" invalid'
        assert type(self.resistor_r2) is int, 'VCC Error: Voltage divider resistor value "resistor_r2" invalid'
        assert type(self.adc_attenuation_db) is float, 'VCC Error: ADC attenuation value "adc_attenuation_db" invalid'

        # ADC channel used for sampling the raw value.
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
            from machine import ADC, Pin
            self.adc = ADC(Pin(int(self.pin[1:])))

        elif self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            from machine import ADC
            self.adc = ADC(id=0)

        else:
            raise NotImplementedError('machine.ADC support is not implemented on this platform')

        # Configure attenuation.
        if self.adc_attenuation_db == 0.0:
            self.adc_atten = ADC.ATTN_0DB
        elif self.adc_attenuation_db == 2.5:
            self.adc_atten = ADC.ATTN_2_5DB
        elif self.adc_attenuation_db == 6.0:
            self.adc_atten = ADC.ATTN_6DB
        elif self.adc_attenuation_db == 11.0:
            self.adc_atten = ADC.ATTN_11DB
        else:
            raise ValueError('ADC attenuation value (adc_attenuation_db) not allowed : {}'.format(self.adc_attenuation_db))

    def read(self):
        """Acquire voltage reading by sampling ADC."""
        # Todo: Make attenuation factor configurable.

        # Sample ADC a few times.
        adc_samples = [0.0] * self.adc_sample_count
        adc_mean = 0.0
        i = 0
        log.debug('Reading voltage level on pin {} with voltage divider {}/{}'.format(self.pin, self.resistor_r1, self.resistor_r2))

        # read samples
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
            self.adc.atten(self.adc_atten)
            irq_state = disable_irq()
            while i < self.adc_sample_count:
                adc_samples[i] = self.adc.read()
                adc_mean += adc_samples[i]
                i += 1
            enable_irq(irq_state)

        elif self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            self.adc.init()
            adc_channel = self.adc.channel(attn=self.adc_atten, pin=self.pin)
            irq_state = disable_irq()
            while i < self.adc_sample_count:
                sample = adc_channel()
                adc_samples[i] = sample
                adc_mean += sample
                i += 1
            enable_irq(irq_state)

        else:
            raise NotImplementedError('Reading the ADC is '
                                      'not implemented on this platform')

        adc_mean /= self.adc_sample_count
        adc_variance = 0.0
        for sample in adc_samples:
            adc_variance += (sample - adc_mean) ** 2
        adc_variance /= (self.adc_sample_count - 1)

        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
            # FIXME: Make this work for vanilla ESP32.
            raw_voltage = 0.0
            mean_voltage = 0.0

        elif self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            raw_voltage = adc_channel.value_to_voltage(4095)
            mean_voltage = adc_channel.value_to_voltage(int(adc_mean))

        mean_variance = (adc_variance * 10 ** 6) // (adc_mean ** 2)

        # log.debug("ADC readings. count=%u:\n%s" %(self.adc_sample_count, str(adc_samples)))
        log.debug("SystemVoltage: Mean of ADC readings (0-4095) = %15.13f" % adc_mean)
        log.debug("SystemVoltage: Mean of ADC voltage readings (0-%dmV) = %15.13f" % (raw_voltage, mean_voltage))
        log.debug("SystemVoltage: Variance of ADC readings = %15.13f" % adc_variance)
        log.debug("SystemVoltage: 10**6*Variance/(Mean**2) of ADC readings = %15.13f" % mean_variance)

        resistor_sum = self.resistor_r1 + self.resistor_r2
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            voltage_millivolt = (adc_channel.value_to_voltage(int(adc_mean))) * resistor_sum / self.resistor_r2
        else:
            # FIXME: Make this work for vanilla ESP32.
            voltage_millivolt = 0.0            
        voltage_volt = voltage_millivolt / 1000.0

        # Shut down ADC channel.
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            adc_channel.deinit()

        log.debug('Voltage level: {}'.format(voltage_volt))

        reading = {self.reading_key: voltage_volt}
        return reading

    def power_off(self):
        """Shut down ADC."""
        log.info('Turning off ADC')
        self.adc.deinit()


class SystemUptime(AbstractSystemSensor):
    """
    Report system uptime.
    https://docs.pycom.io/firmwareapi/micropython/utime.html#utimeticksms
    """

    start_time = time.time()

    def read(self):
        """ """
        now = time.time()
        uptime = time.ticks_ms() / 1000.0
        runtime = now - self.start_time
        reading = {
            'system.time': now,
            'system.uptime': uptime,
            'system.runtime': runtime,
        }
        return reading


class SystemHallSensor:
    """
    - https://forum.pycom.io/topic/3537/internal-hall-sensor-question/2
    - https://github.com/micropython/micropython-esp32/pull/211
    - https://www.esp32.com/viewtopic.php?t=481
    """
    pass


class SystemMemoryPeeker:
    """
    ``machine.mem32[]``
    
    - https://github.com/micropython/micropython-esp32/pull/192#issuecomment-334631994


    """
    pass


class ADC2:
    """
    ::
    
        adc = machine.ADC(machine.Pin(35))
        adc.atten(machine.ADC.ATTN_11DB)
    
    - https://github.com/micropython/micropython-esp32/issues/33
    - https://www.esp32.com/viewtopic.php?t=955

    """
    pass



# -*- coding: utf-8 -*-
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019-2020 Richard Pobering <richard@hiveeyes.org>
# (c) 2019-2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# License: GNU General Public License, Version 3
from binascii import hexlify
from machine import Pin

from terkin import logging
from terkin.sensor.common import AbstractBus

log = logging.getLogger(__name__)
#log.setLevel(logging.DEBUG)


class BusType:
    """  """

    I2C = 'i2c'
    OneWire = 'onewire'


class SensorManager:
    """Manages all buses and sensors."""

    def __init__(self, settings):
        self.sensors = []
        self.buses = {}
        self.settings = settings

    def register_sensor(self, sensor):
        """

        :param sensor: 

        """
        self.sensors.append(sensor)

    def register_bus(self,  bus):
        """

        :param bus: 

        """
        bus_name = bus.name
        log.info('Registering bus "%s"', bus_name)
        self.buses[bus_name] = bus

    def get_bus_by_name(self, name):
        """

        :param name: 

        """
        if name is None:
            log.error('Bus "{}" does not exist'.format(name))
            return

        log.debug('Trying to find bus by name "%s"', name)
        bus = self.buses.get(name)
        log.debug('Found bus by name "%s": %s', name, bus)
        return bus

    def get_bus_by_sensortype(self, type):
        """
        Return the bus for this sensor type.
        """
        for sensor in self.sensors:
            if hasattr(sensor, 'type') and sensor.type == type:
                return sensor.bus

    def get_sensor_by_name(self, name):
        """

        :param name: 

        """
        raise NotImplementedError('"get_sensor_by_name" not implemented yet')

    def get_sensor_by_id(self, id):
        """
        Return sensor with defined id.
        """

        log.info('Trying to find bus by name "%s"', id)
        for sensor in self.sensors:
            log.debug('checking sensor with settings "%s"', sensor.settings)
            if hasattr(sensor, 'settings') and sensor.settings.get('id') == id:
                log.debug('sensor with id {} found'.format(id))
                return sensor
        log.info('sensor with id {} NOT found'.format(id))

    def get_sensor_by_type(self, type):
        """
        Return all sensors filtered by type.
        """
        for sensor in self.sensors:
            if hasattr(sensor, 'type') and sensor.type == type:
                yield sensor

    def get_sensors_by_family(self, family):
        """
        Return all sensors filtered by family.
        """
        for sensor in self.sensors:
            if hasattr(sensor, 'family') and sensor.family == family:
                yield sensor

    def setup_buses(self, buses_settings):
        """Register configured I2C, OneWire and SPI buses.

        :param buses_settings: 

        """

        effective_buses = []
        effective_bus_ids = []
        for bus_settings in buses_settings:

            if bus_settings.get("enabled"):
                effective_buses.append(bus_settings)
                effective_bus_ids.append(bus_settings['id'])

        log.info('Starting buses: %s', effective_bus_ids)

        for bus_settings in effective_buses:
            try:
                self.setup_bus(bus_settings)
            except Exception as ex:
                log.exc(ex, 'Registering bus failed. settings={}'.format(bus_settings))

    def setup_bus(self, bus_settings):
        """

        :param bus_settings:

        """
        bus_family = bus_settings.get('family')

        if bus_family == BusType.OneWire:
            owb = OneWireBus(bus_settings)
            if 'pin_data' in bus_settings:
                owb.register_pin("data", bus_settings['pin_data'])
            owb.start()
            self.register_bus(owb)

        elif bus_family == BusType.I2C:
            i2c = I2CBus(bus_settings)
            if 'pin_sda' in bus_settings:
                i2c.register_pin("sda", bus_settings['pin_sda'])
            if 'pin_scl' in bus_settings:
                i2c.register_pin("scl", bus_settings['pin_scl'])
            i2c.start()
            self.register_bus(i2c)

        else:
            log.error("Invalid bus configuration: %s", bus_settings)

    def power_on(self):
        """Send power-on to all buses and sensors"""
        if self.settings.get('sensors.power_toggle_buses', True):
            self.power_toggle_buses('power_on')
        if self.settings.get('sensors.power_toggle_sensors', True):
            self.power_toggle_sensors('power_on')

    def power_off(self):
        """Send power-off to all buses and sensors"""
        if self.settings.get('sensors.power_toggle_sensors', True):
            self.power_toggle_sensors('power_off')
        if self.settings.get('sensors.power_toggle_buses', True):
            self.power_toggle_buses('power_off')

    def power_toggle_sensors(self, action):
        """

        :param action:

        """
        for sensor in self.sensors:
            sensorname = sensor.__class__.__name__
            if hasattr(sensor, action):
                log.info('Sending {} to sensor {}'.format(action, sensorname))
                try:
                    getattr(sensor, action)()
                except Exception as ex:
                    log.exc(ex, 'Sending {} to sensor {} failed'.format(action, sensorname))

    def power_toggle_buses(self, action):
        """

        :param action:

        """
        for busname, bus in self.buses.items():
            if hasattr(bus, action):
                log.info('Sending {} to bus {}'.format(action, busname))
                try:
                    getattr(bus, action)()
                except Exception as ex:
                    log.exc(ex, 'Sending {} to sensor {} failed'.format(action, busname))

    def start_sensors(self):
        log.info("Starting all sensors")
        for sensor in self.sensors:
            if hasattr(sensor, 'start'):
                try:
                    sensor.start()
                except Exception as ex:
                    log.exc(ex, 'Starting sensor "{}" failed. Reason: {}.'.format(sensor.type, ex))


class OneWireBus(AbstractBus):
    """Initialize the 1-Wire hardware driver and represent as bus object."""

    type = BusType.OneWire

    def start(self):
        """  """
        # Todo: Improve error handling.
        try:

            # Vanilla MicroPython 1.11
            if self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
                pin = Pin(int(self.pins['data'][1:]))
                import onewire_native
                self.adapter = onewire_native.OneWire(pin)

            # Pycom MicroPython 1.9.4
            elif self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
                pin = Pin(self.pins['data'])
                if self.settings.get('driver') == 'native':
                    log.info('Using native 1-Wire driver on Pycom MicroPython')
                    import onewire_native
                    self.adapter = onewire_native.OneWire(pin)
                else:
                    log.info('Using pure-Python 1-Wire driver on Pycom MicroPython')
                    import onewire_python
                    self.adapter = onewire_python.OneWire(pin)

            elif self.platform_info.vendor == self.platform_info.MICROPYTHON.RaspberryPi:
                from terkin.sensor.linux import LinuxSysfsOneWireBus
                sysfs = self.settings['sysfs']
                self.adapter = LinuxSysfsOneWireBus(sysfs)

            else:
                raise NotImplementedError('1-Wire bus support is not implemented on this platform')

            self.scan_devices()
            self.ready = True

        except Exception as ex:
            #log.exc(ex, '1-Wire hardware driver failed')
            #return False
            raise

        return True

    def scan_devices(self):
        """  """

        # The 1-Wire bus sometimes needs a fix when coming back from deep sleep.
        #self.adapter.reset()

        # TODO: Tune this further?
        #time.sleep(1)

        # Scan for 1-Wire devices and remember them.
        # TODO: Refactor things specific to DS18x20 devices elsewhere.
        self.devices = [rom for rom in self.adapter.scan() if rom[0] == 0x10 or rom[0] == 0x28]
        log.info("Found {} 1-Wire (DS18x20) devices: {}".format(len(self.devices), self.get_devices_ascii()))

    def get_devices_ascii(self):
        """  """
        return list(map(self.device_address_ascii, self.devices))

    def serialize(self):
        """  """
        info = super().serialize()
        if 'devices' in info:
            info['devices'] = self.get_devices_ascii()
        return info

    @staticmethod
    def device_address_ascii(address):
        """

        :param address:

        """
        # Compute ASCII representation of device address.
        if isinstance(address, (bytearray, bytes)):
            address = hexlify(address).decode()
        return address.lower()


class I2CBus(AbstractBus):
    """Initialize the I2C hardware driver and represent as bus object."""

    type = BusType.I2C
    frequency = 100000

    def start(self):
        """  """
        # Todo: Improve error handling.
        try:
            if self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
                from machine import I2C
                self.adapter = I2C(self.number, sda=Pin(int(self.pins['sda'][1:])), scl=Pin(int(self.pins['scl'][1:])), freq=self.frequency)

            elif self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
                from machine import I2C
                self.adapter = I2C(self.number, mode=I2C.MASTER, pins=(self.pins['sda'], self.pins['scl']), baudrate=self.frequency)

            elif self.platform_info.vendor == self.platform_info.MICROPYTHON.Odroid:
                from smbus2 import SMBus
                self.adapter = SMBus(self.number)

            elif self.platform_info.vendor == self.platform_info.MICROPYTHON.RaspberryPi:
                import board
                import busio

                def i2c_add_bus(busnum, scl, sda):
                    """
                    Register more I2C buses with Adafruit Blinka.

                    Make Adafruit Blinka learn another I2C bus.
                    Please make sure you define it within /boot/config.txt like::

                    dtoverlay=i2c-gpio,bus=3,i2c_gpio_delay_us=1,i2c_gpio_sda=26,i2c_gpio_scl=20
                    """

                    # Uncache this module, otherwise monkeypatching will fail on subsequent calls.
                    import sys
                    try:
                        del sys.modules['microcontroller.pin']
                    except:
                        pass

                    # Monkeypatch "board.pin.i2cPorts".
                    i2c_port = (busnum, scl, sda)
                    if i2c_port not in board.pin.i2cPorts:
                        board.pin.i2cPorts += (i2c_port,)

                pin_scl = self.pins['scl']
                pin_sda = self.pins['sda']

                # When I2C port pins are defined as Integers, register them first.
                if isinstance(pin_scl, int):
                    i2c_add_bus(self.number, pin_scl, pin_sda)
                    SCL = board.pin.Pin(pin_scl)
                    SDA = board.pin.Pin(pin_sda)

                # When I2C port pins are defined as Strings and start with "board.",
                # they are probably already Pin aliases of Adafruit Blinka.
                elif isinstance(pin_scl, str) and pin_scl.startswith('board.'):
                    SCL = eval(pin_scl)
                    SDA = eval(pin_sda)

                self.adapter = busio.I2C(SCL, SDA)

            else:
                raise NotImplementedError('I2C bus is not implemented on this platform')

            self.just_started = True

            if not self.platform_info.vendor == self.platform_info.MICROPYTHON.Odroid:
                self.scan_devices()

            if self.platform_info.vendor == self.platform_info.MICROPYTHON.Odroid:

                self.devices = self.scan_devices_smbus2()
                log.info("Scan I2C bus via smbus2 for devices...")
                log.info("Found {} I2C devices: {}.".format(len(self.devices), self.devices))

            self.ready = True

        except Exception as ex:
            #log.exc(ex, 'I2C hardware driver failed')
            raise

    def scan_devices(self):
        log.info('Scan I2C with id={} bus for devices...'.format(self.number))
        self.devices = self.adapter.scan()
        # i2c.readfrom(0x76, 5)
        log.info("Found {} I2C devices: {}".format(len(self.devices), self.devices))

    def scan_devices_smbus2(self, start=0x03, end=0x78):
        try:
            list = []
            for i in range(start, end):
                val = 1
                try:
                    self.adapter.read_byte(i)
                except OSError as e:
                    val = e.args[0]
                finally:
                    if val != 5:  # No device
                        if val == 1:
                            res = "Available"
                        elif val == 16:
                            res = "Busy"
                        elif val == 110:
                            res = "Timeout"
                        else:
                            res = "Error code: " + str(val)
                        # print(hex(i) + " -> " + res)
                        if res == 'Available':
                            # print(i)
                            list.append(i)
            return list

        except Exception as exp:
            log.exc(exp, 'scan smbus2 failed')

    def power_on(self):
        """
        Turn on the I2C peripheral after power off.
        """

        # Don't reinitialize device if power on just occurred through initial driver setup.
        if self.just_started:
            self.just_started = False
            return

        # uPy doesn't have deinit so it doesn't need init
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            from machine import I2C
            self.adapter.init(mode=I2C.MASTER, baudrate=self.frequency)

    def power_off(self):
        """
        Turn off the I2C peripheral.

        https://docs.pycom.io/firmwareapi/pycom/machine/i2c.html
        """
        log.info('Turning off I2C bus {}'.format(self.name))
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            self.adapter.deinit()



# -*- coding: utf-8 -*-
# (c) 2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2020 Andreas Motl <andreas.motl@terkin.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create VE.Direct sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = VEDirectSensor(settings=sensor_info)
    return sensor_object


class VEDirectSensor(AbstractSensor):
    """
    About
    =====
    Victron Energy VE.Direct MPPT charge controller sensor component.

    Supported devices
    =================
    - SmartSolar MPPT 100/20
    - SmartSolar MPPT 75/15
    - BlueSolar MPPT 75/15
    - BMV 702 battery monitor

    Resources
    =========
    - https://github.com/karioja/vedirect
    - https://www.victronenergy.com/solar-charge-controllers/smartsolar-mppt-75-10-75-15-100-15-100-20
    - https://www.victronenergy.com/solar-charge-controllers/bluesolar-mppt-150-35
    - https://www.victronenergy.com/battery-monitors/bmv-700
    - https://www.victronenergy.com/live/victronconnect:mppt-solarchargers

    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.device = settings['device']
        self.timeout = 5
        self.driver = None

    def start(self):
        log.info('Initializing sensor "Victron Energy VE.Direct"')

        # Initialize the hardware driver.
        try:

            # MicroPython
            if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
                raise NotImplementedError('VEDirect driver not implemented on MicroPython')

            # CPython
            elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
                from vedirect import Vedirect
                self.driver = Vedirect(serialport=self.device, timeout=self.timeout)

            else:
                raise NotImplementedError('VEDirect driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'VEDirect hardware driver failed')

    def read(self):
        log.info('Reading sensor "Victron Energy VE.Direct"')

        # MicroPython
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
            raise NotImplementedError('VEDirect driver not implemented on MicroPython')

        # CPython
        elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
            data_raw = self.driver.read_data_single()

        data = {}
        for key, value in data_raw.items():
            key = 'vedirect:{}'.format(key)
            data[key] = value

        return data



# -*- coding: utf-8 -*-
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import machine

from terkin import logging
from terkin.util import get_platform_info

log = logging.getLogger(__name__)


class MachineResetCause:
    """
    Machine reset cause and wakeup reason definitions.
    
    Will produce nice log messages like these.
    
    - Coming from power on / hard reset::
    
        {'reset_cause': {'code': 0, 'message': 'PWRON'}, 'wakeup_reason': {'code': 0, 'message': 'PWRON'}}
    
    - Coming from deep sleep::
    
        {'reset_cause': {'code': 3, 'message': 'DEEPSLEEP'}, 'wakeup_reason': {'code': 2, 'message': 'RTC'}}
    
    Works on Pycom MicroPython 1.20.0.rc11, YMMV.
    Pycom MicroPython 1.20.0.rc11 [v1.9.4-0a38f88] on 2019-05-14; FiPy with ESP32, published on 14 May 2019


    """

    reset_cause_candidates = [
        'PWRON',
        'HARD',
        'SOFT',
        'WDT',
        'DEEPSLEEP',
        'BROWN_OUT',
    ]

    wakeup_reason_candidates = [
        'PIN',
        'PWRON',
        'RTC',
        'ULP',
        'WLAN',
    ]

    def __init__(self):
        """
        Resolve reset cause and wakeup reason symbols
        and store in dictionary for lookup through
        ``self.humanize()``.
        """

        self.reset_causes = {}
        self.wakeup_reasons = {}

        for name in self.reset_cause_candidates:
            attribute = name + '_RESET'
            symbol = getattr(machine, attribute, None)
            if symbol is not None:
                self.reset_causes[symbol] = name

        for name in self.wakeup_reason_candidates:
            attribute = name + '_WAKE'
            symbol = getattr(machine, attribute, None)
            if symbol is not None:
                self.wakeup_reasons[symbol] = name

    def humanize(self):
        """
        Yield dictionary containing reset cause and wakeup reason
        suitable for human consumption through an appropriate log message.
        """

        platform_info = get_platform_info()

        reset_cause_magic = machine.reset_cause()
        if platform_info.vendor == platform_info.MICROPYTHON.Pycom:
            wakeup_reason_magic, _ = machine.wake_reason()
        else:
            wakeup_reason_magic = machine.wake_reason()

        log.debug('Reset cause: %s', reset_cause_magic)
        log.debug('Wakeup reason: %s', wakeup_reason_magic)

        reset_cause_label = self.reset_causes.get(reset_cause_magic, 'UNKNOWN')
        wakeup_reason_label = self.wakeup_reasons.get(wakeup_reason_magic, 'UNKNOWN')
        status = {
            'reset_cause': {'code': reset_cause_magic, 'message': reset_cause_label},
            'wakeup_reason': {'code': wakeup_reason_magic, 'message': wakeup_reason_label},
        }
        return status



# -*- coding: utf-8 -*-
# (c) 2019-2020 Richard Pobering <richard@hiveeyes.org>
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.util import get_platform_info

log = logging.getLogger(__name__)


class AbstractSensor:
    """Abstract sensor container, containing meta data as readings."""

    SENSOR_NOT_INITIALIZED = object()

    def __init__(self, settings=None):

        self.settings = settings or {}
        self.type = self.settings.get('type')

        self.name = None
        self.family = None
        self.driver = None

        """
        e.g. Multiple onewire sensors are address indexed on a bus.
        """
        self.address = None
        self.bus = None
        self.parameter = {}
        self.pins = {}

    def start(self):
        """ """
        raise NotImplementedError("Must be implemented in sensor driver")

    def set_address(self, address):
        """

        :param address:

        """
        self.address = address

    def register_pin(self, name, pin):
        """

        :param name:
        :param pin:

        """
        self.pins[name] = pin

    def register_parameter(self, name, parameter):
        """

        :param name:
        :param parameter:

        """
        self.parameter[name] = parameter

    def acquire_bus(self, bus):
        """
        Associate sensor object with bus object.
        :param bus:
        """

        # Skip sensor if associated bus is disabled in configuration.
        if bus is None:
            sensor_type = self.settings.get('type', 'unknown').lower()
            sensor_id = self.settings.get('id', self.settings.get('key', sensor_type))
            sensor_bus = self.settings.get('bus')
            log.warning('Bus {} for sensor with id={} and type={} is disabled, '
                        'skipping registration'.format(sensor_bus, sensor_id, sensor_type))

        self.bus = bus

    def ensure_bus(self):
        if self.bus is None or not self.bus.ready:
            raise KeyError('Bus "{}" not ready for "{}"'.format(self.bus.name, self.__class__.__name__))

    def read(self):
        """ """
        raise NotImplementedError()

    def format_fieldname(self, name, address=None, channel=None):
        """

        :param name:
        :param address:

        """
        parts = [name, channel, address, self.bus.name]
        parts = [part for part in parts if part is not None]
        fieldname = '.'.join(parts)
        return fieldname

    def serialize(self):
        """ """
        return dict(serialize_som(self.__dict__, stringify=['bus']))


class AbstractBus:
    """A blueprint for all bus objects."""

    type = None

    def __init__(self, settings):
        """
        convention <type>:<index>
        """
        self.settings = settings
        self.number = self.settings['number']

        # HAL driver adapter.
        self.adapter = None

        # TODO: Publish found 1-Wire devices to MQTT bus and HTTP API.
        self.devices = []
        self.pins = {}

        # Indicate whether the bus driver is ready.
        self.ready = False

        # Indicate whether the bus driver just has been started.
        self.just_started = None

        # Reference to platform information.
        self.platform_info = get_platform_info()

    @property
    def name(self):
        """ """
        return str(self.type) + ":" + str(self.number)

    def register_pin(self, name, pin):
        """

        :param name:
        :param pin:

        """
        self.pins[name] = pin

    def serialize(self):
        """ """
        info = dict(serialize_som(self.__dict__))
        # FIXME: Why is that?
        info.update({'name': self.name, 'type': self.type})
        return info


def serialize_som(thing, stringify=None):
    """Serialize the sensor object model to a representation
    suitable to be served for the device API.

    :param thing:
    :param stringify:  (Default value = None)

    """
    stringify = stringify or []

    if isinstance(thing, list):
        hm = []
        for item in thing:
            hm.append(serialize_som(item))
        return hm

    elif isinstance(thing, dict):
        newthing = {}
        for key, value in thing.items():
            if key in stringify:
                newthing[key] = str(value)
            else:
                newthing[key] = serialize_som(value)
        return newthing

    elif isinstance(thing, (str, int, float, bool, set)) or type(thing) is type(None):
        return thing

    else:
        if hasattr(thing, 'serialize'):
            thing = thing.serialize()
        else:
            thing = str(thing)
        return thing



# -*- coding: utf-8 -*-
# (c) 2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2020 Andreas Motl <andreas.motl@terkin.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create PiUSV+ sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = PiUSVSensor(settings=sensor_info)
    return sensor_object


class PiUSVSensor(AbstractSensor):
    """
    About
    =====
    PiUSV+ sensor component.

    Supported devices
    =================
    - PiUSV+ v. 1.0

    Resources
    =========
    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        self.driver = None

    def start(self):
        log.info('Initializing sensor "PiUSV+"')

        # Initialize the hardware driver.
        try:

            # CPython
            if platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
                from rpi_piusv import PiUSV
                import smbus
                bus = smbus.SMBus(1)

                # Patch status definition for old firmwares. Unklar!
                PiUSV.status_definition['usb_power'] = 0x02
                PiUSV.status_definition['external_power'] = 0x01

                self.driver = PiUSV(bus)

                #print('Firmware version:', self.driver.read_firmware_version())
                # print('Data:', self.driver.read())

            else:
                raise NotImplementedError('PiUSV+ driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'PiUSV+ hardware driver failed')

    def read(self):

        log.info('Reading sensor "PiUSV+"')

        # CPython
        if platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
            data_raw = self.driver.read()

            data = {}
            for key, value in data_raw.items():
                key = 'piusv:{}'.format(key)
                data[key] = value

            return data
        else:
            return none
